(* ::Package:: *)

(* ::Title:: *)
(*ConvexPolytope`*)


BeginPackage["ConvexPolytope`",{"Vector`","Numeric`","List`"}];


(* ::Subtitle:: *)
(*Names*)


ConvexPolytope;
ConvexPolytopeDual;
ConvexPolytopeHalfSpaces;
ConvexPolytopeIntersectingQ;
ConvexPolytopeIntersection;
ConvexPolytopeChop;
ConvexPolytopeData;
ConvexPolytopeDimension;
ConvexPolytopeEmbed;
ConvexPolytopeFaceCounts;
ConvexPolytopeFigure;
ConvexPolytopeGraphics;
ConvexPolytopeGraphics3D;
ConvexPolytopeMeasure;
ConvexPolytopeProject;
ConvexPolytopeRandomPoint;
ConvexPolytopeRegionMemberQ;
ConvexPolytopeSlice;
ConvexPolytopeTabView;
ConvexPolytopeTransform;
ConvexPolytopeTranslate;


HalfSpacesAffineSpan;
HalfSpacesBoundedQ;
HalfSpacesCoordinateBounds;
HalfSpacesFeasibleQ;
HalfSpacesHyperPlanes;
HalfSpacesObjectiveBound;
HalfSpacesObjectiveVertex;
HalfSpacesRandomVertex;
HalfSpacesReduce;
HalfSpacesTransfer;


(* ::Subtitle::Closed:: *)
(*Private`*)


Begin["`Private`"];


(* ::Subtitle:: *)
(*ConvexPolytope*)


(* ::Subsection::Closed:: *)
(*ConvexPolytope | init*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope*)


ConvexPolytope::usage = 
	"A ConvexPolytope[data] represents a convex polytope where data is an Association which contains information for the vertices, faces, and facet normals.";


ConvexPolytope[] := ConvexPolytope[ConvexPolytopeData[]]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeData*)


ConvexPolytopeData::usage = 
	"ConvexPolytopeData[] creates an Association which defines the ConvexPolytope data format.";


ConvexPolytopeData[] := 
	Association[
			Rule["Dimension",-1],
			Rule["EmbeddingDimension",-1],
			Rule["VertexCoordinates",{}],
			Rule["FaceVertexIndices",Association[]],
			Rule["SimplexVertexIndices",{}],
			Rule["AffineSpaceBasis",{}],
			Rule["HyperPlaneNormals",{}],
			Rule["HyperPlaneBounds",{}]
		]


(* ::Subsection::Closed:: *)
(*ConvexPolytope | vertices*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope*)


ConvexPolytope[{}] := ConvexPolytope[]


ConvexPolytope[ points_?MatrixQ ] :=
	Module[
		{
			affineDimension,
			convexPolytope
		},
		affineDimension = VectorAffineSpanDimension[points];
		If[Equal[affineDimension,0],
			convexPolytope = ConvexPolytopeDimension0[points];
			Return[convexPolytope];
		];
		If[Equal[affineDimension,1],
			convexPolytope = ConvexPolytopeDimension1[points];
			Return[convexPolytope];
		];
		If[Greater[affineDimension,1],
			convexPolytope = ConvexPolytopeDimensionN[points];
			Return[convexPolytope];
		];
	];


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeDimension0*)


ConvexPolytopeDimension0[pointCoordinates_] := 
	Module[{
			embeddingDimension,
			vertex,
			convexPolytopeData,
			convexPolytope
		},
		embeddingDimension = VectorDimension[pointCoordinates];
		vertex = Part[pointCoordinates,1];
		convexPolytopeData = ConvexPolytopeData[];
		convexPolytopeData["Dimension"] = 0;
		convexPolytopeData["EmbeddingDimension"] = embeddingDimension;
		convexPolytopeData["VertexCoordinates"] = {vertex};
		convexPolytopeData["FaceVertexIndices"] = Association[Rule[0,{{1}}]];
		convexPolytopeData["SimplexVertexIndices"] = {{1}};
		convexPolytopeData["AffineSpaceBasis"] = {};
		convexPolytopeData["HyperPlaneNormals"] = IdentityMatrix[embeddingDimension];
		convexPolytopeData["HyperPlaneBounds"] = vertex;
		convexPolytopeData["HalfSpaceNormals"] = {};
		convexPolytopeData["HalfSpaceBounds"] = {};
		convexPolytope = ConvexPolytope[convexPolytopeData];
		Return[convexPolytope];
]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeDimension1*)


ConvexPolytopeDimension1[pointCoordinates_] := 
	Module[{
			points,
			embeddingDimension,
			vertex1,
			vertex2,
			normalVector,
			convexPolytopeData,
			convexPolytope
		},
		points = pointCoordinates;		
		embeddingDimension = VectorDimension[points];
		(* lexicograph ordering gives extremal points *)
		vertex1 = Part[points,First[Ordering[points,+1]]];
		vertex2 = Part[points,First[Ordering[points,-1]]];
		normalVector = NChop[Normalize[vertex2-vertex1]];
		convexPolytopeData = ConvexPolytopeData[];
		convexPolytopeData["Dimension"] = 1;
		convexPolytopeData["EmbeddingDimension"] = embeddingDimension;
		convexPolytopeData["VertexCoordinates"] = {vertex1,vertex2};
		convexPolytopeData["FaceVertexIndices"] = <|0->{{1},{2}},1->{{1,2}}|>;
		convexPolytopeData["SimplexVertexIndices"] = {{1,2}};
		convexPolytopeData["AffineSpaceBasis"] = {normalVector};
		convexPolytopeData["HyperPlaneNormals"] = NChop[Orthogonalize[NullSpace[{normalVector}]]];
		convexPolytopeData["HyperPlaneBounds"] = VectorProject[convexPolytopeData["HyperPlaneNormals"],vertex1];
		convexPolytopeData["HalfSpaceNormals"] = {-normalVector,normalVector};
		convexPolytopeData["HalfSpaceBounds"] = {Dot[-normalVector,vertex1],Dot[normalVector,vertex2]};
		convexPolytope = ConvexPolytope[convexPolytopeData];
		Return[convexPolytope];
]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeDimensionN*)


ConvexPolytopeDimensionN[pointsInput_] := 
	Module[{
			points,
			affineSpace,
			simplex,
			centroid,
			simplexes,
			vertexPointIndices,
			vertices,
			facets,
			normals,
			bounds,
			faces,
			normalSignatures,
			rulesPointToVertexIndices,
			convexPolytopeData,
			convexPolytope
			},
		(* points *)
			points = pointsInput;
			(*points = N[points];*)
			affineSpace = VectorAffineSpace[N[points]];
		(* project to affine space *)
			{facets,normals,bounds,simplexes} = ConvexPolytopeSimplicialFacets[points];
		(* convert to vertex indices *)
			{vertices, facets, simplexes} = ConvexPolytopeReindexPointsToVertices[points, facets, simplexes];
		(* merge faces *)
			{faces, normals, bounds} = ConvexPolytopeMergeSimplicialFacets[{facets,normals,bounds}];
		(* chop numerical values *)
			normals = NChop[normals];
			bounds = NChop[bounds];
		(* output polytope *)
			convexPolytopeData = ConvexPolytopeData[];
			convexPolytopeData["Dimension"] = affineSpace["Dimension"];
			convexPolytopeData["EmbeddingDimension"] = affineSpace["EmbeddingDimension"];
			convexPolytopeData["VertexCoordinates"] = vertices;
			convexPolytopeData["FaceVertexIndices"] = faces;
			convexPolytopeData["SimplexVertexIndices"] = simplexes;
			convexPolytopeData["AffineSpaceBasis"] = affineSpace["Basis"];
			convexPolytopeData["HyperPlaneNormals"] = affineSpace["NullSpaceBasis"];
			convexPolytopeData["HyperPlaneBounds"] = VectorProject[vertices[[1]],convexPolytopeData["HyperPlaneNormals"]];
			convexPolytopeData["HalfSpaceNormals"] = normals;
			convexPolytopeData["HalfSpaceBounds"] = bounds;
			convexPolytope = ConvexPolytope[convexPolytopeData];
		Return[ConvexPolytope[convexPolytopeData]];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeSimplicialFacets*)


ConvexPolytopeSimplicialFacets[pointCoordinatesInput_] :=
	Module[
		{
			pointCoordinates,
				pointIndices,
				embeddingDimension,
				centroid,
				affineSpace,
				affineSpaceDimension,
				affineNullSpaceDimension,
				affineSpaceBasis,
				affineNullSpaceBasis,
				affineSpaceBasePoint,
			simplexVertexPointIndices,
				simplexVertexCoordinates,
				lookupSimplexFacets,
				simplexes,
			vertexPointIndices,
				vertices,
			facets,
				normals,
				bounds,
			remPointIndices,
				remFacetPointHeights,
				remPointFacetHeights,
				remPointQs,
				remPointCoordinates,
			facetIndex,
				normal,
				bound,
				remHeights,
				maxHeight,
				maxPointIndices,
				newVertexPointIndex,
				newVertex,
				newVertexFacetSignatures,
				visibleFacetBs,
				visibleFacets,
				newSimplexes,
				newFacets,
				newNormals,
				newBounds
		},
		(* input *)
			pointCoordinates = pointCoordinatesInput;
			pointCoordinates = N[pointCoordinates];
			pointIndices = ListIndices[pointCoordinates];
		(* affine space *)
			affineSpace = VectorAffineSpace[pointCoordinates];
			affineSpaceBasis = affineSpace["Basis"];
			affineSpaceDimension = affineSpace["Dimension"];
			affineNullSpaceBasis = affineSpace["NullSpaceBasis"];
			affineNullSpaceDimension = affineSpace["NullSpaceDimension"];
			affineSpaceBasePoint = affineSpace["BasePoint"];
		(* project to affine space *)
			pointCoordinates = VectorProject[pointCoordinates,affineSpaceBasis];
		(* initial affine simplex *)
			simplexVertexPointIndices = ConvexPolytopeExtremalSimplexVertexIndices[pointCoordinates];
			simplexVertexCoordinates = Part[pointCoordinates,simplexVertexPointIndices];
		(* recenter: initial simplex centroid \[RightTeeArrow] origin *)
			centroid = VectorCentroid[simplexVertexCoordinates];
			pointCoordinates = VectorTranslate[pointCoordinates,Minus[centroid]];
		(* simplex *)
			lookupSimplexFacets = Subsets[Range[Plus[1,affineSpaceDimension]],{affineSpaceDimension}];
		(* initial simplex *)
			simplexes = {simplexVertexPointIndices};
			vertexPointIndices = simplexVertexPointIndices;
			vertices = Part[pointCoordinates,vertexPointIndices];
		(* initial facets *)
			facets = Parts[simplexVertexPointIndices,lookupSimplexFacets];
			normals = VectorNormalize[Map[SimplexNormal,Parts[pointCoordinates,facets]]];
			bounds = NChop[MapThread[Dot,{normals,pointCoordinates[[facets[[All,1]]]]}]];
			(* orient normals *)
				normals *= Sign[bounds];
				bounds *= Sign[bounds];
		(* build polytope simplices *)
			remPointIndices = Complement[pointIndices,simplexVertexPointIndices];
			While[Unequal[{},remPointIndices],
				(* remove interior remaining points *)
					remFacetPointHeights = NChop[Subtract[Dot[normals,Transpose[Part[pointCoordinates,remPointIndices]]],bounds]];
					If[NoneTrue[Sign[remFacetPointHeights],EqualTo[1],2],Break[];];
					remPointFacetHeights = Transpose[remFacetPointHeights];
					remPointQs = Map[AnyTrue[Positive],Transpose[remFacetPointHeights]];
					remPointIndices = Pick[remPointIndices,remPointQs];
					remPointFacetHeights = Pick[remPointFacetHeights,remPointQs];
					remFacetPointHeights = Transpose[remPointFacetHeights];
				(* choose new vertex (extremal) *)
					facetIndex = First[Ordering[Map[Max,remFacetPointHeights],-1]];
					normal = normals[[facetIndex]];
					bound = bounds[[facetIndex]];
					remPointCoordinates = pointCoordinates[[remPointIndices]];
					remHeights = NRound[remPointCoordinates.normal-bound];
					maxHeight = Max[remHeights];
					maxPointIndices = Pick[remPointIndices,remHeights,maxHeight];
					newVertexPointIndex = Part[maxPointIndices,First[Ordering[Part[pointCoordinates,maxPointIndices],Minus[1]]]];
					newVertex = pointCoordinates[[newVertexPointIndex]];
				(* add new vertex to polytope *)
					vertexPointIndices = Append[vertexPointIndices,newVertexPointIndex];
					vertices = Append[vertices,newVertex];
				(* extrude visible facets to new simplexes *)
					visibleFacetBs = NPositiveBoole[Subtract[Dot[normals,newVertex],bounds]];
					visibleFacets = Pick[facets,visibleFacetBs,1];
				(* keep non-visible {facets,normals,bounds} *)
					facets = Pick[facets,visibleFacetBs,0];
					normals = Pick[normals,visibleFacetBs,0];
					bounds = Pick[bounds,visibleFacetBs,0];
				(* get new facets from new simplexes - remove duplicated(\[DoubleLeftRightArrow]interior) facets *)
					newSimplexes = ArrayPad[visibleFacets,{{0,0},{0,1}},newVertexPointIndex];
					newFacets = Flatten[Table[Parts[newSimplex,Rest[lookupSimplexFacets]],{newSimplex,newSimplexes}],1];
					newFacets = Sort[Map[Sort,newFacets]];
					newFacets = ListNonDuplicates[newFacets];
				(* add new simplexes to polytope *)
					newSimplexes = Map[Sort,newSimplexes];
					simplexes = Join[simplexes,newSimplexes];
				(* compute new affine normals, bounds *)
					newNormals = Map[SimplexNormal,Parts[pointCoordinates,newFacets]];
					newNormals = VectorNormalize[newNormals];
					newBounds = Dot[newNormals,newVertex];
					(* flip inward pointing normals *)
						newNormals *= Sign[newBounds];
						newBounds = Abs[newBounds];
				(* add new facets to polytope *)
					facets = Flatten[{facets,newFacets},1];
					normals = Flatten[{normals,newNormals},1];
					bounds = Flatten[{bounds,newBounds},1];
				];
			(* recenter to (simplex centroid) \[Function] origin *)
				bounds = Plus[bounds,Dot[normals,centroid]];
			(* embed normals *)
				normals = VectorEmbed[normals,affineSpaceBasis];
			Return[{facets,normals,bounds,simplexes}];
		]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeExtremalSimplexVertexIndices*)


ConvexPolytopeExtremalSimplexVertexIndices[ pointsInput_ ] := 
	Module[
		{
			points,
			embeddingDimension,
			vectors,
			pivot,
			nullSpaceBasis,
			extremalVertexIndex,
			extremalVectorIndex,
			extremalVector,
			simplexVertexIndices
		},
		(* input *)
			points = N[pointsInput];
			embeddingDimension = VectorDimension[points];
		(* init *)
			nullSpaceBasis = IdentityMatrix[embeddingDimension];
			simplexVertexIndices = {};
		(* initial vertex *)
			pivot = First[points];
			extremalVertexIndex = First[Ordering[Minus[VectorSquaredNorm[VectorTranslate[points,Minus[pivot]]]],1]];
			pivot = Part[points,extremalVertexIndex];
			extremalVertexIndex = First[Ordering[Minus[VectorSquaredNorm[VectorTranslate[points,Minus[pivot]]]],1]];
			pivot = Part[points,extremalVertexIndex];
			vectors = VectorTranslate[points,Minus[pivot]];
		(* record initial vertex *)
			simplexVertexIndices = {extremalVertexIndex};
		Do[
			extremalVectorIndex = First[Ordering[Minus[VectorSquaredNorm[vectors]],1]];
			extremalVector = Part[vectors,extremalVectorIndex];
			If[Equal[Chop[Norm[extremalVector]],0],Break[];];
			simplexVertexIndices = Append[simplexVertexIndices,extremalVectorIndex];
			nullSpaceBasis = NullSpace[{extremalVector}];
			If[Equal[nullSpaceBasis,{}],Break[];];
			vectors = VectorProject[vectors,nullSpaceBasis];
		,{dimension,1,embeddingDimension}];
		Return[simplexVertexIndices];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeReindexPointsToVertices*)


ConvexPolytopeReindexPointsToVertices[pointCoordinates_, facetPointIndices_, simplexPointIndices_] := 
	Module[{
			vertexPointIndices,
			vertexIndices,
			vertexCoordinates,
			lookupPointToVertexIndices,
			facetVertexIndices,
			simplexVertexIndices
		},
		vertexPointIndices = Union[Flatten[facetPointIndices]];
		vertexIndices = ListIndices[vertexPointIndices];
		lookupPointToVertexIndices = ConstantArray[0,Max[vertexPointIndices]];
		lookupPointToVertexIndices[[vertexPointIndices]] = vertexIndices;
		facetVertexIndices = Parts[lookupPointToVertexIndices,facetPointIndices];
		simplexVertexIndices = Parts[lookupPointToVertexIndices,simplexPointIndices];
		vertexCoordinates = Part[pointCoordinates,vertexPointIndices];
		Return[{vertexCoordinates, facetVertexIndices, simplexVertexIndices}];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeMergeSimplicialFacets*)


ConvexPolytopeMergeSimplicialFacets[{simplicialFacets_, simplicialFacetNormals_, simplicialFacetBounds_}] := 
	Module[
		{
			polytopeDimension,
			facetComplexes,
			simplicialFacetIndexBins,
			mergedFacets,
			vertexFacetIncidences,
			GetIncidentMergedFacetIndices,
			faces
		},
		polytopeDimension = Length[First[simplicialFacets]];
		simplicialFacetIndexBins = GatherIndices[NRound[simplicialFacetNormals]];
		facetComplexes = Parts[simplicialFacets,simplicialFacetIndexBins];
		facetComplexes = Map[Sort,facetComplexes,2];
		facetComplexes = SortBy[facetComplexes,Flatten];
		mergedFacets = Apply[Union,facetComplexes,1];
		faces = Association[Table[Rule[n,{}],{n,0,polytopeDimension-1}]];
		faces[polytopeDimension-1] = mergedFacets;
		vertexFacetIncidences = IndexSubsetMemberships[mergedFacets];
		GetIncidentMergedFacetIndices = Function[face,Apply[Intersection,Part[vertexFacetIncidences,face]]];
		Module[
			{
				complexes,
				complexBoundarySimplexes,
				facetBoundarySimplexes,
				subSimplexes,
				subComplexes,
				subFaces
			},
			complexes = facetComplexes;
			Do[
				complexBoundarySimplexes = 
					Table[
						ListNonDuplicates[Map[Sort,Flatten[Subsets[Transpose[complexes[[complexIndex]]],{faceDimension}],{3,1}]]]
					,{complexIndex,Length[complexes]}];
				subSimplexes = Union[Flatten[complexBoundarySimplexes,1]];
				subComplexes = GatherBy[subSimplexes,GetIncidentMergedFacetIndices];
				subComplexes = SortBy[subComplexes,Flatten];
				subFaces = Apply[Union,subComplexes,1];
					faces[faceDimension-1] = subFaces;
				(* for next iteration *)
				complexes = subComplexes;
			,{faceDimension,polytopeDimension-1,1,-1}];
		];
		Return[
			{
			faces,
			Part[simplicialFacetNormals,simplicialFacetIndexBins[[All,1]]],
			Part[simplicialFacetBounds,simplicialFacetIndexBins[[All,1]]]
			}
		];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeSimplicialFaceComplexes*)


ConvexPolytopeSimplicialFaceComplexes[ nFacesInput_ ] := 
	Module[{
			nFaces,
			nFaceComplexes,
			faces,
			face,
			boundaryFaceQs,
			boundaryComplexes,
			boundaryFaces,
			cornerVertexIndex,
			boundarySimplexes,
			faceSimplexes,
			faceComplexes,
			faceDimension
		},
		nFaces = nFacesInput;
		nFaceComplexes = nFaces;
		nFaceComplexes[0] = Map[List,nFaceComplexes[0],{-2}];
		nFaceComplexes[1] = Map[List,nFaceComplexes[1],{-2}];
		Do[
			nFaceComplexes[faceDimension] = 
			Table[
				face = nFaces[faceDimension][[faceIndex]];
				If[Equal[Length[face],Plus[1,faceDimension]],
					faceSimplexes = List[face];
					,
					cornerVertexIndex = Min[face];
					boundaryFaceQs = nFaces[faceDimension-1] // Map[Function[boundaryFace,And[Not[MemberQ[boundaryFace,cornerVertexIndex]],ContainsOnly[boundaryFace,face]]]];
					boundaryComplexes = Pick[nFaceComplexes[faceDimension-1],boundaryFaceQs];
					boundarySimplexes = Union[Apply[Join,boundaryComplexes]];
					faceSimplexes = ArrayPad[boundarySimplexes,{{0,0},{0,1}},cornerVertexIndex];
					faceSimplexes = Sort[Map[Sort,faceSimplexes]];
				];
				faceSimplexes
			,{faceIndex,Length[nFaces[faceDimension]]}];
		,{faceDimension,Complement[Keys[nFaces],{0,1}]}];
		Return[nFaceComplexes];
		]


(* ::Subsection::Closed:: *)
(*ConvexPolytope | HalfSpaces*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeHalfSpaces*)


ConvexPolytopeHalfSpaces::NotFeasible = 
	"halfspaces do not intersect";


ConvexPolytopeHalfSpaces::NotBounded = 
	"halfspace intersection not bounded";


ConvexPolytopeHalfSpaces[{normalsInput_, boundsInput_}] := 
	Module[{
			embeddingDimension,
			
			halfSpaceNormals,
			halfSpaceBounds,
			
			hyperPlaneNormals,
			hyperPlaneBounds,
			
			halfSpaceNormalVectorSpace,
			halfSpaceNormalEmbeddedQ,
			
			feasibleQ,
			boundedQ,
			interiorPoint,
			
			affineSpace,
			recenterQ,
			affineEmbeddedQ,
			
			dualPoints,
			dualFacets,
			dualNormals,
			dualBounds,
			dualSimplexes,
			
			dualFacetIndices,
			vertices,
			convexPolytope
		},
		
		(* input *)
			embeddingDimension = VectorDimension[normalsInput];
			halfSpaceNormals = NChop[N[normalsInput]];
			halfSpaceBounds = NChop[N[boundsInput]];
		
		(* recude half-spaces *)
			{halfSpaceNormals,halfSpaceBounds} = HalfSpacesReduce[{halfSpaceNormals,halfSpaceBounds}];

		(* project to span(halfspace normals) *)
			halfSpaceNormalVectorSpace = VectorSpace[halfSpaceNormals];
			halfSpaceNormalEmbeddedQ = Less[halfSpaceNormalVectorSpace["Dimension"],embeddingDimension];
			If[halfSpaceNormalEmbeddedQ,
				halfSpaceNormals = VectorProject[halfSpaceNormals,halfSpaceNormalVectorSpace["Basis"]];];
			
		(* check feasibility *)
			If[Not[HalfSpacesFeasibleQ[{halfSpaceNormals,halfSpaceBounds}]], 
				Message[ConvexPolytopeHalfSpaces::NotFeasible]; 
				Return[ConvexPolytope[]];
			];
		
		(* check bounded *)
			If[Not[HalfSpacesBoundedQ[{halfSpaceNormals,halfSpaceBounds}]],
				Message[ConvexPolytopeHalfSpaces::NotBounded]; 
				Return[ConvexPolytope[]];
			];
		
		(* affine span *)
			affineSpace = HalfSpacesAffineSpan[{halfSpaceNormals,halfSpaceBounds}];
		
		(* case: dimension \[Equal] 0 *)
			If[Equal[affineSpace["Dimension"],0], 
				Return[ConvexPolytope[{affineSpace["BasePoint"]}]];];
					
		(* recenter interiorPoint \[Function] Origin *)
			interiorPoint = affineSpace["BasePoint"];
			halfSpaceBounds = NChop[Subtract[halfSpaceBounds,Dot[halfSpaceNormals,interiorPoint]]];
		
		(* transfer halfspaces to affine space *)
			affineEmbeddedQ = Less[affineSpace["Dimension"],embeddingDimension];
			If[affineEmbeddedQ,
				{halfSpaceNormals,halfSpaceBounds} = HalfSpacesTransfer[{halfSpaceNormals,halfSpaceBounds},affineSpace["Basis"]];];
			
		(* get dual facets - spherical inversion *)
			dualPoints = Divide[halfSpaceNormals,halfSpaceBounds];
			{dualFacets,dualNormals,dualBounds,dualSimplexes} = ConvexPolytopeSimplicialFacets[dualPoints];
		
		(* select one simplex for each facet (after grouping) *)
			dualFacetIndices = Part[Values[PositionIndex[NRound[dualNormals]]],All,1];
			dualFacets = Part[dualFacets,dualFacetIndices];
		
		(* vertices *)
			vertices = Table[LinearSolve[Part[halfSpaceNormals,dualFacet],Part[halfSpaceBounds,dualFacet]],{dualFacet,dualFacets}];
			
		(* map vertices back into original vector/affine space *)
			If[affineEmbeddedQ, vertices = VectorEmbed[vertices,affineSpace["Basis"]];];
			vertices = VectorTranslate[vertices,interiorPoint];
			If[halfSpaceNormalEmbeddedQ, vertices = VectorEmbed[vertices,halfSpaceNormalVectorSpace["Basis"]]];
		(* chop numerical values *)
			vertices  = NChop[vertices];			
		(* polytope *)
			convexPolytope = ConvexPolytope[vertices];
		(* output *)
			Return[convexPolytope];
	]


(* ::Subsection:: *)
(*ConvexPolytope | intersection*)


(* ::Subsubsection:: *)
(*ConvexPolytopeIntersection*)


ConvexPolytopeIntersection[ p1_ConvexPolytope, p2_ConvexPolytope ] := 
	Module[{aff1,aff2,aff,n,b,p},
		aff1 = VectorAffineSpace[NChop[VectorCentroid[p1["VertexCoordinates"]]],p1["AffineSpaceBasis"]];
		aff2 = VectorAffineSpace[NChop[VectorCentroid[p2["VertexCoordinates"]]],p2["AffineSpaceBasis"]];
		aff = VectorAffineSpaceIntersection[aff1,aff2];
		If[Equal[aff["Dimension"],Minus[1]],Return[ConvexPolytope[]];];
		If[Equal[aff["Dimension"],Minus[1]],Return[ConvexPolytope[{aff["BasePoint"]}]];];
		n = Join[p1["HalfSpaceNormals"],p2["HalfSpaceNormals"]];
		b = Join[p1["HalfSpaceBounds"],p2["HalfSpaceBounds"]];
		(* re-center halfspaces at interior point *)
		b = Subtract[b,Dot[n,aff["BasePoint"]]];
		{n,b} = HalfSpacesTransfer[{n,b},aff["Basis"]];
		n = NChop[n];
		b = NChop[b];
		p = ConvexPolytopeHalfSpaces[{n,b}];
		If[NonNegative[p["Dimension"]],
			p = ConvexPolytopeEmbed[p,aff["Basis"]];
			p = ConvexPolytopeTranslate[p,aff["BasePoint"]];
		];
		Return[p];
	]


(* ::Subsubsection:: *)
(*ConvexPolytopeIntersectingQ*)


ConvexPolytopeIntersectingQ[ p1_ConvexPolytope, p2_ConvexPolytope ] := 
	Module[{
			normals1, bounds1, signedBounds1,
			normals2, bounds2, signedBounds2,
			point1,
			point2,
			point1New,
			point2New,
			intersectingQ
		},
		normals1 = p1["HalfSpaceNormals"] // N;
		normals2 = p2["HalfSpaceNormals"] // N;
		bounds1 = p1["HalfSpaceBounds"] // N;
		bounds2 = p2["HalfSpaceBounds"] // N;
		signedBounds1 = Thread[{bounds1,-1}];
		signedBounds2 = Thread[{bounds2,-1}];
		
		point1 = p1["VertexCoordinates"][[1]] // N;
		point2 = p2["VertexCoordinates"][[1]] // N;
		
		intersectingQ = False;
		While[Not[intersectingQ],
			If[Or[VectorBoundedQ[point1, normals2, bounds2],VectorBoundedQ[point2, normals1, bounds1]],
				intersectingQ = True;
				Break[];
			];
			point1New = LinearProgramming[ Subtract[point2,point1], normals1, signedBounds1, Minus[Infinity]];
			point2New = LinearProgramming[ Subtract[point1,point2], normals2, signedBounds2, Minus[Infinity]];
			If[And[NEqual[point1,point1New],NEqual[point2,point2New]], 
				intersectingQ = False;
				Break[];
			];
		];
		Return[intersectingQ];
	];		


(* ::Subsection::Closed:: *)
(*ConvexPolytope | methods*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | methods*)


ConvexPolytope[data_][methodName_, args___] := ConvexPolytopeKeyMaster[ConvexPolytope[data],methodName,args]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeKeyMaster*)


ConvexPolytopeKeyMaster::usage = 
	"ConvexPolytopeKeyMaster[polytope, method, (args\[Ellipsis])] is a handler for the ConvexPolytope data type which determines and calls the appropriate function as indicated by the given method.";


ConvexPolytopeKeyMaster[ConvexPolytope[data_], keyInput_, args___] := 
	Module[{key = keyInput},
		
		If[{"VertexCoordinates", "Vertices", "vertices", "v", "vc"} // MemberQ[key],
			Return[data["VertexCoordinates"]];];
		
		If[{"FaceVertexIndices", "fvi"} // MemberQ[key],
			Return[ConvexPolytopeFaces[ConvexPolytope[data],args]];];
		
		If[{"FaceVertexCoordiantes", "FaceVertices", "FaceCoordinates", "Faces", "faces", "f", "fv", "fvc", "fc"} // MemberQ[key],
			Return[Parts[data["VertexCoordinates"],ConvexPolytopeFaces[ConvexPolytope[data],args]]];];
		
		If[{"FacetVertexCoordinates", "FacetCoordinates", "FacetVertices", "Facets", "facets"} // MemberQ[key],
			Return[Parts[data["VertexCoordinates"],ConvexPolytopeFaces[ConvexPolytope[data],-1]]];];
			
		If[{"HalfSpaceNormals", "FacetNormals", "Normals", "normals", "n", "hn"} // MemberQ[key],
			Return[data["HalfSpaceNormals"]];];
		
		If[{"HalfSpaceBounds", "FacetBounds", "Bounds", "bounds", "b", "hb"} // MemberQ[key],
			Return[data["HalfSpaceBounds"]];];
		
		If[{"EmbeddingDimension", "VectorDimension", "AmbientDimension", "vdim", "embdim"} // MemberQ[key],
			Return[data["EmbeddingDimension"]];];
		
		If[{"Dimension", "AffineDimension", "dimension", "affdim", "dim"} // MemberQ[key],
			Return[data["Dimension"]];];
		
		If[{"FullDimensionQ", "DimensionFullQ"} // MemberQ[key],
			Return[Equal[data["Dimension"], data["EmbeddingDimension"]]];];
		
		If[{"AffineSpaceBasis", "AffineSpace", "AffineBasis", "affspace", "aff", "a"} // MemberQ[key],
			Return[data["AffineSpaceBasis"]];];
		
		If[{"AffineNullSpace", "affnull", "null", "perp", "\[Perpendicular]"} // MemberQ[key],
			Return[data["HyperPlaneNormals"]];];
		
		If[{"Centroid", "CentroidCoordinates", "centroid", "intp", "mean", "c"} // MemberQ[key],
			Return[Mean[data["VertexCoordinates"]]];];
		
		If[{"RandomPoint", "randp", "rand", "rp", "r"} // MemberQ[key],
			Return[ConvexPolytopeRandomPoint[ConvexPolytope[data], args]];];
		
		If[{"Translate", "translate"} // MemberQ[key],
			Return[ConvexPolytopeTranslate[ConvexPolytope[data], args]];];
		
		If[{"Embed", "embed"} // MemberQ[key],
			Return[ConvexPolytopeEmbed[ConvexPolytope[data], args]];];
			
		If[{"Project", "project", "Projection", "projection", "Proj", "proj"} // MemberQ[key],
			Return[ConvexPolytopeProject[ConvexPolytope[data], args]];];
			
		If[{"FaceCounts", "FaceTally", "f#", "fcount"} // MemberQ[key],
			Return[ConvexPolytopeFaceCounts[ConvexPolytope[data], args]];];
		
		If[{"Slice", "slice"} // MemberQ[key],
			Return[ConvexPolytopeSlice[ConvexPolytope[data], args]];];
		
		If[{"Figure", "fig"} // MemberQ[key],
			Return[ConvexPolytopeFigure[ConvexPolytope[data], args]];];
			
		If[{"TabView", "Tab", "tab"} // MemberQ[key],
			Return[ConvexPolytopeTabView[ConvexPolytope[data]]];];
		
		If[{"Graphics3D", "g3d", "g3D", "g3", "gfx"} // MemberQ[key],
			Return[ConvexPolytopeGraphics3D[ConvexPolytope[data], args]];];
			
		If[{"Graphics", "g2d", "g2D", "g2"} // MemberQ[key],
			Return[ConvexPolytopeGraphics[ConvexPolytope[data], args]];];
		
		If[{"RegionMemberQ", "InteriorQ"} // MemberQ[key],
			Return[ConvexPolytopeRegionMemberQ[ConvexPolytope[data], args]];];
			
		If[{"Measure", "measure", "Volume", "volume", "vol"} // MemberQ[key],
			Return[ConvexPolytopeMeasure[ConvexPolytope[data], args]];];
		
		Return["---< foo.exe >---"];
	]


(* ::Subsection::Closed:: *)
(*ConvexPolytope | attributes*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeFaces*)


ConvexPolytopeFaces[ConvexPolytope[data_]] := data["FaceVertexIndices"]
ConvexPolytopeFaces[ConvexPolytope[data_], faceDimensionInput_Integer] := 
	Module[{
			polytopeDimension,
			faceDimension,
			faceCodimension,
			result
		},
		result = {};
		faceDimension = faceDimensionInput;
		polytopeDimension = data["Dimension"];
		If[LessEqual[Minus[polytopeDimension],faceDimension,-1],
			faceCodimension = Abs[faceDimension];
			faceDimension = Subtract[polytopeDimension,faceCodimension];
			result = data["FaceVertexIndices",faceDimension];
		];
		If[LessEqual[0,faceDimension,Subtract[polytopeDimension,1]],
			result = data["FaceVertexIndices",faceDimension];
		];
		Return[result];
	];


ConvexPolytopeFaces[ConvexPolytope[data_], faceDimensionsInput_List] := 
	Map[Function[faceDimension,ConvexPolytopeFaces[ConvexPolytope[data],faceDimension]],faceDimensionsInput,{-1}]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeFaceCounts*)


ConvexPolytopeFaceCounts::usage = 
	"PolytopeFaceCounts[polytope] outputs the tallies of face vertex counts for each face dimension.";


ConvexPolytopeFaceCounts[ConvexPolytope[data_],faceDimensionsInput_:Null] := 
	Module[
		{
			faceDimensions,
			faceDimension,
			faceCodimension,
			polytopeDimension,
			faceVertexIndices,
			nfaces,
			tally,
			tallies,
			labels,
			label,
			result
		},
		faceDimensions = faceDimensionsInput;
		polytopeDimension = data["Dimension"];
		faceVertexIndices = data["FaceVertexIndices"];
		
		If[SameQ[faceDimensions,Null],
			faceDimensions = Range[0,Subtract[Length[faceVertexIndices],1]];
		];
		
		If[Not[ListQ[faceDimensions]],
			faceDimensions = List[faceDimensions];
		];
		
				
		tallies = 
		Table[
			tally = "\[EmptySet]";
			
			If[LessEqual[0,faceDimension,polytopeDimension],
				tally = MatrixForm[Tally[Map[Length,faceVertexIndices[[Plus[1,faceDimension]]]]]];
			];
			
			If[LessEqual[Minus[polytopeDimension],faceDimension,Minus[1]],
				faceCodimension = Abs[faceDimension];
				faceDimension = Subtract[polytopeDimension,faceCodimension];
				tally = MatrixForm[Tally[Map[Length,faceVertexIndices[[Plus[1,faceDimension]]]]]]
			];
			
			tally
			
		,{faceDimension,faceDimensions}];
		
		labels = 
			Table[
				label = "\[EmptySet]";
				
				If[LessEqual[0,faceDimension,polytopeDimension],
					label = ToString[faceDimension]<>"-faces";
				];
				
				If[LessEqual[Minus[polytopeDimension],faceDimension,Minus[1]],
					faceCodimension = Abs[faceDimension];
					faceDimension = Subtract[polytopeDimension,faceCodimension];
					label = ToString[faceDimension]<>"-faces";
				];
				
				label
		,{faceDimension,faceDimensions}];
		
		result = Grid[Transpose[MapThread[List,{labels,tallies}]], Dividers->1];
		
		Return[result];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeMeasure*)


ConvexPolytopeMeasure[ConvexPolytope[data_]] := 
	If[data["Dimension"]==0,1,
	If[Equal[{},data["SimplexVertexIndices"]],
		Undefined,
		Total[Map[SimplexMeasure,Parts[data["VertexCoordinates"],data["SimplexVertexIndices"]]]]
	]]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeDimension*)


ConvexPolytopeDimension[ConvexPolytope[data_]] := Lookup[data,"Dimension"]


(* ::Subsection::Closed:: *)
(*ConvexPolytope | print*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeFigure*)


ConvexPolytopeFigure[ConvexPolytope[dataInput_],figureLabel_:"Polytope"] := 
	Module[{data,faceVertexIndices,faces,figure},
		data = dataInput;
		faceVertexIndices = data["FaceVertexIndices"];
		faces = Grid[Table[List["dim="<>ToString[faceDimension],Shallow[faceVertexIndices[[faceDimension+1]]]],{faceDimension,0,Length[faceVertexIndices]-1}],Rule[Dividers,Hue[0,0,0]],Rule[Alignment,{{Right,Left}}]];
		data = Map[Shallow,data];
		data["FaceVertexIndices"] = faces;
		figure = Grid[Apply[List,Normal[data],1],Rule[Dividers,1], Rule[Alignment,{{Right,Left}}]];
		Return[figure];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeOpenerView*)


ConvexPolytopeOpenerView // ClearAll;
ConvexPolytopeOpenerView[ConvexPolytope[data_]] := 
	Module[
		{
			polytopeData,
			faceVertexIndices,
			facesTabView,
			dataOpenerView
		},
		polytopeData = data;
		dataOpenerView = polytopeData;
		dataOpenerView = MapAt[Shallow,dataOpenerView,{{"VertexCoordinates"},{"HalfSpaceNormals"},{"HalfSpaceBounds"},{"FaceVertexIndices",All}}];
		dataOpenerView["FaceVertexIndices"] = 
			OpenerView[{
				"FacesVertexIndices",
				Column@Table[
					OpenerView[{
						ToString[faceDimension]<>"-faces",
						dataOpenerView["FaceVertexIndices",faceDimension]
					}]
				,{faceDimension,0,polytopeData["Dimension"]-1}]
			}];
		dataOpenerView = OpenerView[{"ConvexPolytope",Column[Map[OpenerView@*Apply[List],Normal[dataOpenerView]]]}];
		Return[dataOpenerView];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeTabView*)


ConvexPolytopeTabView[ConvexPolytope[dataInput_]] := 
	Module[{data,faceVertexIndices,facesTabView,dataTabView},
		data = dataInput;
		faceVertexIndices = data["FaceVertexIndices"];
		facesTabView = 
			TabView[
				Table[Rule["dim="<>ToString[faceDimension]<>": ",Short[faceVertexIndices[[faceDimension+1]]]],{faceDimension,0,Length[faceVertexIndices]-1}]
			,ControlPlacement->Left];
		data = Map[Shallow,data];
		data["FaceVertexIndices"] = facesTabView;
		dataTabView = TabView[data,ControlPlacement->Left,ImageSize->512];
		Return[dataTabView];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | TabView*)


ConvexPolytope/: TabView[ConvexPolytope[data_],options___] := TabView[data,options]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | MenuView*)


ConvexPolytope/: MenuView[ConvexPolytope[data_],options___] := MenuView[data,options]


(* ::Subsection::Closed:: *)
(*ConvexPolytope | Graphics*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | Simplex*)


ConvexPolytope /: Simplex[dilation_:1][polytope_ConvexPolytope] := ConvexPolytopeGraphics3D[polytope, Simplex, dilation]


ConvexPolytope /: Simplex[polytope_ConvexPolytope, dilation_:1] := ConvexPolytopeGraphics3D[polytope, Simplex, dilation]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | Polygon*)


ConvexPolytope /: Polygon[polytope_ConvexPolytope, dilation_:1] := ConvexPolytopeGraphics3D[polytope, Polygon, dilation]


ConvexPolytope /: Polygon[dilation_:1][polytope_ConvexPolytope] := ConvexPolytopeGraphics3D[polytope, Polygon, dilation]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | Point*)


ConvexPolytope /: Point[polytope_ConvexPolytope] := ConvexPolytopeGraphics3D[polytope, Point]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | Line*)


ConvexPolytope /: Line[polytope_ConvexPolytope] := ConvexPolytopeGraphics3D[polytope, Line]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | Sphere*)


ConvexPolytope /: Sphere[polytope_ConvexPolytope, radius_:0.1] := ConvexPolytopeGraphics3D[polytope, Sphere, radius]


ConvexPolytope /: Sphere[radius_:0.1][polytope_ConvexPolytope] := ConvexPolytopeGraphics3D[polytope, Sphere, radius]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope | Tube*)


ConvexPolytope /: Tube[polytope_ConvexPolytope, radius_:0.05, dilation_:1] := ConvexPolytopeGraphics3D[polytope, Tube, radius, dilation]


ConvexPolytope /: Tube[radius_:0.05, dilation_:1][polytope_ConvexPolytope] := ConvexPolytopeGraphics3D[polytope, Tube, radius, dilation]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeGraphics3D*)


ConvexPolytopeGraphics3D[ConvexPolytope[polytope_], Point] := 
	Module[{vertices3D,graphics3D},
		If[Less[polytope["Dimension"],0],Return[{}];];
		vertices3D = polytope["VertexCoordinates"].IdentityMatrix[{polytope["EmbeddingDimension"],3}];
		graphics3D = Point[vertices3D];
		Return[graphics3D];
	]


ConvexPolytopeGraphics3D[ConvexPolytope[polytope_], Sphere, radius_:1] := 
	Module[{vertices3D,graphics3D},
		If[Less[polytope["Dimension"],0],Return[{}];];
		vertices3D = polytope["VertexCoordinates"].IdentityMatrix[{polytope["EmbeddingDimension"],3}];
		graphics3D = Sphere[vertices3D,radius];
		Return[graphics3D];
	]


ConvexPolytopeGraphics3D[ConvexPolytope[polytope_], Line] := 
	Module[{vertices3D,edges3D,graphics3D},
		If[Less[polytope["Dimension"],1],Return[{}];];
		vertices3D = polytope["VertexCoordinates"].IdentityMatrix[{polytope["EmbeddingDimension"],3}];
		edges3D = Parts[vertices3D,polytope["FaceVertexIndices",1]];
		graphics3D = Line[edges3D];
		Return[graphics3D];
	]


ConvexPolytopeGraphics3D[ConvexPolytope[polytope_], Tube, radius_:.05, dilation_:1] := 
	Module[{vertices3D,edges3D,graphics3D},
		If[Less[polytope["Dimension"],1],Return[{}];];
		vertices3D = polytope["VertexCoordinates"].IdentityMatrix[{polytope["EmbeddingDimension"],3}];
		edges3D = Parts[vertices3D,polytope["FaceVertexIndices",1]];
		If[Unequal[1,dilation],
			edges3D = Map[VectorDilate[dilation],edges3D];];
		graphics3D = Tube[edges3D,radius];
		Return[graphics3D];
	]


ConvexPolytopeGraphics3D[ConvexPolytope[polytope_], Polygon, dilation_:1] := 
	Module[{vertices3D,polygons3D,graphics3D},
		If[Less[polytope["Dimension"],2],Return[{}];];
		If[Equal[polytope["Dimension"],2],Return[Polygon[PolygonVertexSort[polytope["VertexCoordinates"].IdentityMatrix[{polytope["EmbeddingDimension"],3}]]]];];
		vertices3D = polytope["VertexCoordinates"].IdentityMatrix[{polytope["EmbeddingDimension"],3}];
		polygons3D = Parts[vertices3D,polytope["FaceVertexIndices",2]];
		polygons3D = Map[PolygonVertexSort,polygons3D];
		If[Unequal[1,dilation],
			polygons3D = Map[VectorDilate[dilation],polygons3D];];
		graphics3D = Polygon[polygons3D];
		Return[graphics3D];
	]


ConvexPolytopeGraphics3D[ConvexPolytope[polytope_], Simplex, dilation_:1] := 
	Module[{vertices3D,polygons3D,graphics3D},
		If[Less[polytope["Dimension"],0],Return[{}];];
		If[Equal[polytope["Dimension"],0],Return[Simplex[{polytope["VertexCoordinates"][[1]]}]];];
		vertices3D = polytope["VertexCoordinates"].IdentityMatrix[{polytope["EmbeddingDimension"],3}];
		polygons3D = Parts[vertices3D,polytope["SimplexVertexIndices"]];
		If[Unequal[dilation,1],
			polygons3D = Map[VectorDilate[dilation],polygons3D];];
		graphics3D = Map[Simplex,polygons3D];
		Return[graphics3D];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeGraphics*)


ConvexPolytopeGraphics[ConvexPolytope[data_],Point] := 
	Module[{vertices2D,graphics2D},
		vertices2D = data["VertexCoordinates"].IdentityMatrix[{data["EmbeddingDimension"],2}];
		graphics2D = Point[vertices2D];
		Return[graphics2D];
	]


ConvexPolytopeGraphics[ConvexPolytope[data_],Disk,radius_:0.1] := 
	Module[{vertices2D,graphics2D},
		vertices2D = data["VertexCoordinates"].IdentityMatrix[{data["EmbeddingDimension"],2}];
		graphics2D = Table[Disk[vertex,radius],{vertex,vertices2D}];
		Return[graphics2D];
	]


ConvexPolytopeGraphics[ConvexPolytope[data_],Line] := 
	Module[{vertices3D,edges3D,graphics3D},
		vertices3D = data["VertexCoordinates"].IdentityMatrix[{data["EmbeddingDimension"],2}];
		edges3D = Parts[vertices3D,data["FaceVertexIndices"][[2]]];
		graphics3D = Line[edges3D];
		Return[graphics3D];
	]


ConvexPolytopeGraphics[ConvexPolytope[data_],Tube,width_:0.05] := 
	Module[{vertices,edges,graphics2D,normals,vectors,corners},
		vertices = data["VertexCoordinates"].IdentityMatrix[{data["EmbeddingDimension"],2}];
		edges = Parts[vertices,data["FaceVertexIndices"][[2]]];
		vectors = Map[First@*Differences,edges];
		normals = VectorNormalize[Map[Cross,vectors]];
		corners = Table[edges[[index,1]] - normals[[index]]*width,{index,Length[edges]}];
		graphics2D = Table[Parallelogram[corners[[index]],{vectors[[index]],2*width*normals[[index]]}],{index,Length[edges]}];
		Return[graphics2D];
	]


ConvexPolytopeGraphics[ConvexPolytope[data_],Polygon,dilation_:1] := 
	Module[{vertices2D,polygon2D,graphics2D},
		vertices2D = data["VertexCoordinates"].IdentityMatrix[{data["EmbeddingDimension"],2}];
		polygon2D = PolygonOrient[vertices2D];
		If[Unequal[1,dilation],polygon2D = VectorDilate[polygon2D,dilation];];
		graphics2D = Polygon[polygon2D];
		Return[graphics2D];
	]


(* ::Subsection::Closed:: *)
(*ConvexPolytope | points*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeRandomPoint*)


ConvexPolytopeRandomPoint::usage = 
	"ConvexPolytopeRandomPoint[ convexPolytope ] returns a random point in the interior of convexPolytope.
	ConvexPolytopeRandomPoint[ convexPolytope, n ] returns a list of n random points in the interior of convexPolytope.";


ConvexPolytopeRandomPoint[ ConvexPolytope[data_], arrayDimensions_:Nothing ] := 
	Module[{ vertices, weights, randomPoints, randomPoint },
		vertices = data["VertexCoordinates"];
		weights = RandomReal[1,{arrayDimensions,Length[vertices]}]^data["Dimension"];
		weights /= weights.ConstantArray[1,Length[vertices]];
		randomPoints = Dot[weights,vertices];
		Return[randomPoints];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeRegionMemberQ*)


ConvexPolytopeRegionMemberQ[ConvexPolytope[data_], points_] := VectorBoundedQ[points,data["HalfSpaceNormals"],data["HalfSpaceBounds"]]
ConvexPolytopeRegionMemberQ[ConvexPolytope[data_]][points_] := ConvexPolytopeRegionMemberQ[ConvexPolytope[data], points]


(* ::Subsection::Closed:: *)
(*ConvexPolytope | operations*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeDual*)


ConvexPolytopeDual::usage = 
	"ConvexPolytopeDual[polytope] returns the dual of polytope via reciprocation about the unit sphere (centered at the polytope centroid).";


ConvexPolytopeDual[polytope_] := 
	Module[{
			faceDimension,
			faceIndex,
			faceFacetIndices,
			dualPolytope,
			polytopeRecenterQ,
			vertices,
			polytopeEmbedQ,
			dualBounds,
			pointFacetIncidences,
			origin,
			facetIndices,
			pointFacetAdj,
			pointFacetIndices,
			points,
			centroid,
			normals,
			bounds,
			facets,
			faces,
			affineDimension,
			embeddingDimension,
			dimension,
			affineSubspace,
			dualVertices,
			dualNormals,
			facetPoints,
			dualFaces
		},
		points = polytope[[1,"VertexCoordinates"]];
		dimension = polytope[[1,"Dimension"]];
		faces = polytope[[1,"FaceVertexIndices"]];
		normals = polytope[[1,"HalfSpaceNormals"]];
		bounds = polytope[[1,"HalfSpaceBounds"]];
		points = points // N // Chop;
		normals = normals // N // Chop;
		bounds = bounds // N // Chop;
		facets = faces[dimension-1];
		centroid = Mean[points] // Chop;
		polytopeRecenterQ = AnyTrue[bounds,NonPositive];
		If[polytopeRecenterQ,
			points = VectorTranslate[points,Minus[centroid]];
			bounds = Subtract[bounds,Dot[normals,centroid]];
		];
		embeddingDimension = VectorDimension[points];
		polytopeEmbedQ = Less[dimension,embeddingDimension];
		
		If[polytopeEmbedQ,
			affineSubspace = VectorAffineSpace[points]["Basis"] // Orthogonalize // Chop;
			normals = VectorProject[normals,affineSubspace];
			points = VectorProject[points,affineSubspace];
		];
		
		facetIndices = Range[Length[facets]];
		vertices = points[[Flatten[faces[[1]]]]];
		dualVertices = Table[LinearSolve[points[[facet]],ConstantArray[1,Length[facet]]],{facet,facets}];
		pointFacetAdj = SparseArray[Rule[Flatten[MapIndexed[Composition[Tuples,List],facets],1],1]];
		pointFacetIndices = Map[Function[pointSubsetAdj,Pick[facetIndices,pointSubsetAdj,1]],pointFacetAdj];		
		pointFacetIndices = Round[pointFacetIndices];
		Monitor[
			dualFaces = Table[Apply[Intersection,pointFacetIndices[[faces[[faceDimension,faceIndex]]]]],{faceDimension,dimension},{faceIndex,Length[faces[[faceDimension]]]}];
			,Column[{ProgressIndicator[faceDimension,{1,dimension}],ProgressIndicator[faceIndex,{1,Length[faces[[faceDimension]]]}]}],1];
		dualFaces = Reverse[dualFaces];
		dualFaces = Association[Table[Rule[faceDimension,dualFaces[[faceDimension+1]]],{faceDimension,0,dimension-1}]];
		dualNormals = VectorNormalize[vertices];
		dualBounds = MapThread[Dot,{dualNormals,dualVertices[[dualFaces[[dimension,All,1]]]]}];
		
		If[polytopeEmbedQ,
			dualVertices = VectorEmbed[dualVertices,affineSubspace];
			dualNormals = VectorEmbed[dualNormals,affineSubspace];
			];
		
		If[polytopeRecenterQ,
			dualVertices = VectorTranslate[dualVertices,centroid];
			dualBounds += Dot[dualNormals,centroid];
			];

		dualVertices = NChop[dualVertices];
		dualNormals = NChop[dualNormals];
		dualBounds = NChop[dualBounds];
		dualPolytope = polytope;
		dualPolytope[[1,"VertexCoordinates"]] = dualVertices;
		dualPolytope[[1,"FaceVertexIndices"]] = dualFaces;
		dualPolytope[[1,"HalfSpaceNormals"]] = dualNormals;
		dualPolytope[[1,"HalfSpaceBounds"]] = dualBounds;
		dualPolytope[[1,"Dimension"]] = dimension;
		dualPolytope[[1,"SimplexVertexIndices"]] = {};
		Return[dualPolytope];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeTranslate*)


ConvexPolytopeTranslate::usage = 
	"PolytopeTranslate[polytope,vector] translates the polytope by vector, updating the polytope points and bounds.";


ConvexPolytopeTranslate // ClearAll;
ConvexPolytopeTranslate[ConvexPolytope[dataInput_], vector_] := 
	Module[{
			data,
			polytope,
			vertexCoordinates,
			pointCoordinates,
			halfSpaceNormals,
			halfSpaceBounds,
			HyperPlaneNormals,
			HyperPlaneBounds
		},
		data = dataInput;
		vertexCoordinates = data["VertexCoordinates"];
		pointCoordinates = data["PointCoordinates"];
		
		halfSpaceNormals = data["HalfSpaceNormals"];
		halfSpaceBounds = data["HalfSpaceBounds"];
		
		HyperPlaneNormals = data["HyperPlaneNormals"];
		HyperPlaneBounds = data["HyperPlaneBounds"];
		
		vertexCoordinates = Transpose[Plus[Transpose[vertexCoordinates],vector]];
		pointCoordinates = Transpose[Plus[Transpose[vertexCoordinates],vector]];
		halfSpaceBounds = Plus[halfSpaceBounds,Dot[halfSpaceNormals,vector]];
		
		If[Unequal[{},HyperPlaneNormals],
			HyperPlaneBounds = Plus[HyperPlaneBounds,Dot[HyperPlaneNormals,vector]];];
		
		data["VertexCoordinates"] = vertexCoordinates // NChop;
		data["HalfSpaceBounds"] = halfSpaceBounds // NChop;
		data["HyperPlaneBounds"] = HyperPlaneBounds // NChop;
		
		polytope = ConvexPolytope[data];
		Return[polytope];
	]
	
ConvexPolytopeTranslate[vector_][polytope_] := ConvexPolytopeTranslate[polytope,vector]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeEmbed*)


ConvexPolytopeEmbed::usage = 
	"ConvexPolytopeProject[polytope,subspace] returns the polytope as embedded into the subspace (assuming the polytope and subspace dimensions are equal). PolytopeEmbed[subspace] represents an operator that can be applied to a polytope.";


ConvexPolytopeEmbed[ConvexPolytope[dataInput_], basis_] := 
Module[{
		data,
		embeddingDimension,
		basisNullSpace,
		vertexCoordinates,
		halfSpaceNormals,
		affineSpaceBasis,
		hyperPlaneNormals,
		hyperPlaneBounds,
		polytopeResult
	},
	data = dataInput;
	If[data["Dimension"]==-1, Return[ConvexPolytope[]];];
	
	embeddingDimension = VectorDimension[basis];
	basisNullSpace = Orthogonalize[NullSpace[N[basis]]];
	
	vertexCoordinates = data["VertexCoordinates"];
	halfSpaceNormals = data["HalfSpaceNormals"];
	
	affineSpaceBasis = data["AffineSpaceBasis"];
	hyperPlaneNormals = data["HyperPlaneNormals"];
	hyperPlaneBounds = data["HyperPlaneBounds"];
	
	vertexCoordinates = VectorEmbed[vertexCoordinates,basis];
	If[Unequal[{},halfSpaceNormals], halfSpaceNormals = Dot[halfSpaceNormals,basis];];
	If[Unequal[{},affineSpaceBasis], affineSpaceBasis = Dot[affineSpaceBasis,basis];];
	If[Unequal[{},hyperPlaneNormals], hyperPlaneNormals = Dot[hyperPlaneNormals,basis];];
	
	If[Unequal[{},basisNullSpace],
		hyperPlaneNormals = Join[hyperPlaneNormals,basisNullSpace];
		hyperPlaneBounds = Join[hyperPlaneBounds,ConstantArray[0,Length[basisNullSpace]]];
	];
	
	data["VertexCoordinates"] = vertexCoordinates // NChop;
	data["HalfSpaceNormals"] = halfSpaceNormals // NChop;
	data["AffineSpaceBasis"] = affineSpaceBasis // NChop;
	data["HyperPlaneNormals"] = hyperPlaneNormals // NChop;
	data["HyperPlaneBounds"] = hyperPlaneBounds // NChop;
	data["EmbeddingDimension"] = embeddingDimension;
	
	polytopeResult = ConvexPolytope[data];
	
	Return[polytopeResult];
]


ConvexPolytopeEmbed[ basis_ ][ ConvexPolytope[data_] ] := ConvexPolytopeEmbed[ ConvexPolytope[data], basis ]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeProject*)


ConvexPolytopeProject::usage = 
	"ConvexPolytopeProject[polytope,subspace] returns the polytope as projected to the subspace. PolytopeProject[subspace] represents an operator that can be applied to a polytope.";


ConvexPolytopeProject[ConvexPolytope[data_], basis_] := ConvexPolytope[NChop[VectorProject[data["VertexCoordinates"],basis]]]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeTransform*)


ConvexPolytopeTransform::usage = "PolytopeTransform[polytope,matrix] returns the polytope with vertices and normals transformed by matrix.";


ConvexPolytopeTransform[basis_][polytope_ConvexPolytope] := ConvexPolytopeTransform[polytope, basis]


ConvexPolytopeTransform[ConvexPolytope[polytopeDataInput_], basis_] := 
	Module[
		{
			polytope,
			embeddingDimension,
			basisNullSpace,
			vertexCoordinates,
			pointCoordinates,
			halfSpaceNormals,
			affineSpaceBasis,
			hyperPlaneNormals,
			polytopeResult
		},
		polytope = polytopeDataInput;
			
		vertexCoordinates = polytope["VertexCoordinates"];
		halfSpaceNormals = polytope["HalfSpaceNormals"];
		affineSpaceBasis = polytope["AffineSpaceBasis"];
		hyperPlaneNormals = polytope["HyperPlaneNormals"];
		
		vertexCoordinates = VectorTransform[vertexCoordinates,basis];
		halfSpaceNormals = VectorTransform[halfSpaceNormals,basis];
		affineSpaceBasis = VectorTransform[affineSpaceBasis,basis];
		hyperPlaneNormals = VectorTransform[hyperPlaneNormals,basis];
		
		polytope["VertexCoordinates"] = vertexCoordinates;
		polytope["HalfSpaceNormals"] = halfSpaceNormals // NChop;
		polytope["AffineSpaceBasis"] = affineSpaceBasis // NChop;
		polytope["HyperPlaneNormals"] = hyperPlaneNormals // NChop;
		
		polytopeResult = ConvexPolytope[polytope];
		
		Return[polytopeResult];
	]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeEdgeSlice*)


ConvexPolytopeEdgeSlice[{p1_,p2_},{v1_,v2_}] := 
	Switch[
		Sign[{v1,v2}],
		{-1,-1},{},
		{-1,0},{p2},
		{-1,1},{(p2*v1 - p1*v2)/(v1-v2)},
		{1,-1},{(p2*v1 - p1*v2)/(v1-v2)},
		{1,1},{},
		{1,0},{p2},
		{0,-1},{p1},
		{0,0},{p1,p2},
		{0,1},{p1}
	];


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeSlice*)


ConvexPolytopeSlice::usage = 
	"ConvexPolytopeSlice[polytope,normal,bound] returns the hyper-planar slice of polytope.";


ConvexPolytopeSlice[polytope_,normal_,value_] := 
	Module[{
			vertices,
			verticesN,
			vertexValues,
			vertexSigns,
			
			edgeVertexIndices,
			edgeIndices,
			edgeVertices,
			edgeVertexValues,
			edgeVertexSigns,
			
			slicedEdgeIndices,
			slicedEdgeVertexIndices,
			slicedEdgeVertices,
			slicedEdgeVertexValues,
			slicedEdgeVertexSigns,
			p1,p2,v1,v2,
			slicedVertices,
			sliceVertices,
			slicePolytope
		},
		vertices = polytope[[1,"VertexCoordinates"]];
		vertices = vertices // N;
		vertexValues = Chop[Subtract[Dot[vertices,normal],value]];
		vertexSigns = Sign[vertexValues];
		sliceVertices = Pick[vertices,vertexSigns,0];
		edgeVertexIndices = polytope[[1,"FaceVertexIndices"]][[2]];
		edgeIndices = Range[Length[edgeVertexIndices]];
		edgeVertexSigns = Parts[vertexSigns,edgeVertexIndices];
		slicedEdgeIndices = Pick[edgeIndices,Map[Sort,edgeVertexSigns],{-1,1}];
		slicedEdgeVertexIndices = Part[edgeVertexIndices,slicedEdgeIndices];
			
		If[Positive[Length[slicedEdgeIndices]],
			slicedEdgeVertices = Parts[vertices,slicedEdgeVertexIndices];
			slicedEdgeVertexValues = Parts[vertexValues,slicedEdgeVertexIndices];
			{v1,v2} = Transpose[slicedEdgeVertexValues];
			{p1,p2} = Flatten[slicedEdgeVertices,{2}];
			slicedVertices = Divide[Chop[Subtract[Times[p2,v1],Times[p1,v2]]],Subtract[v1,v2]];
			,
			slicedVertices = {};
		];
		sliceVertices = Join[sliceVertices,slicedVertices];
		sliceVertices = sliceVertices // NDeleteDuplicates;
		slicePolytope = ConvexPolytope[sliceVertices];
		Return[slicePolytope];
	]


ConvexPolytopeSlice[normal_,value_][polytope_] := ConvexPolytopeSlice[polytope,normal,value]


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeEdgeChop*)


ConvexPolytopeEdgeChop[{p1_,p2_},{v1_,v2_}] := 
	Switch[
		Sign[{v1,v2}],
		{-1,-1},{p1,p2},
		{-1,0},{p1,p2},
		{-1,1},{p1,(p2 v1-p1 v2)/(v1-v2)},
		{1,-1},{(p2 v1-p1 v2)/(v1-v2),p2},
		{1,1},Nothing,
		{1,0},{p2,p2},
		{0,-1},{p1,p2},
		{0,0},{p1,p2},
		{0,1},{p1,p1}
	];


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeChop*)


ConvexPolytopeChop::usage = 
	"ConvexPolytopeChop[polytope,normal,bound] returns the polytope as intersection with the half-space given by the normal and bound.";


ConvexPolytopeChop[polytopeInput_,normalsInput_,boundsInput_] := 
	Module[{polytope,polytopeDimension,slicePolytope,polytopeNormals,polytopeNormalIndices,polytopeBounds,normals,normalIndices,bounds,edges,edgeValues,slicePoints,chopPolytope,normalIndex,normal,bound,points,newEdges,slicePointNormalIndices,slicePointIndices},
		polytope = polytopeInput;
		normals = normalsInput;
		normalIndices = ListIndices[normals];
		bounds = boundsInput;
		polytopeNormals = polytope["normals"];
		polytopeBounds = polytope["bounds"];
		polytopeDimension = polytope["dimension"];
		edges = Parts[polytope["vertices"],polytope["faces",1]];
		edges = edges // Map[NumericalSort];
		chopPolytope = ConvexPolytope[];
		Do[
			normal = normals[[normalIndex]];
			bound = bounds[[normalIndex]];
			edgeValues = Chop[Subtract[Dot[edges,normal],bound]];
			edges = MapThread[ConvexPolytopeEdgeChop,{edges,edgeValues}];
			If[normalIndex == Length[normals],Break[];];
			If[Equal[{},edges],Break[];];
			points = Flatten[edges,1] // NDeleteDuplicates;
			slicePoints = VectorMaximalByNormal[points,normal];			
			newEdges = Select[Table[VectorMaximalByNormal[slicePoints,N@normal],{normal,polytopeNormals}],EqualTo[2]@*Length];
			newEdges = newEdges // Map[NumericalSort];
			edges = Join[edges,newEdges] // NDeleteDuplicates;
		,{normalIndex,Length[normals]}];
		points = Flatten[edges,1] // NDeleteDuplicates;
		chopPolytope = ConvexPolytope[points];
		Return[chopPolytope];
	]


ConvexPolytopeChop[normals_,values_][polytope_] := ConvexPolytopeChop[polytope,normals,values]


(* ::Subtitle::Closed:: *)
(*HalfSpaces*)


(* ::Subsubsection::Closed:: *)
(*HalfSpacesReduce*)


HalfSpacesReduce[{normalsInput_?MatrixQ,boundsInput_?VectorQ}] := 
	Module[{
			normals,
			bounds
		},
		(* input *)
			normals = NChop[N[normalsInput]];
			bounds = NChop[N[boundsInput]];
			
		(* gather parallel normals *)
			Module[{indexBins},
				indexBins = GatherIndices[NRound[VectorNormalize[normals]]];
				normals = Part[normals,Part[indexBins,All,1]];
				bounds = Map[Min,Parts[bounds,indexBins]];
			];
			
		(* sort *)
			Module[{ordering},
				ordering = Ordering[normals];
				normals = Part[normals,ordering];
				bounds = Part[bounds,ordering];
			];

		Return[{normals,bounds}];
	]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesTransfer*)


(* assume origin is interior *)
HalfSpacesTransfer[{nInput_,bInput_},basisInput_] := 
	Module[{basis,n,b,nQs},
		(* input *)
			basis = N[basisInput];
			n = N[nInput];
			b = N[bInput];
		(* project normals to basis *)
			n = NChop[VectorProject[n,basis]];
		(* select (\[Eta],\[Beta])'s with non-trivial projection *)
			nQs = NPositiveBoole[VectorNorm[n]];
			n = Pick[n,nQs,1];
			b = Pick[b,nQs,1];
		(* re-normalize *)
			b = Divide[b,VectorNorm[n]];
			n = VectorNormalize[n];
		(* reduce *)
			{n,b} = HalfSpacesReduce[{n,b}];
		Return[{n,b}];
	]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesHyperPlanes*)


HalfSpacesHyperPlanes[{halfSpaceNormalsInput_,halfSpaceBoundsInput_}] := 
	Module[{
			halfSpaceNormals,
			halfSpaceBounds,
			halfSpaceNormalBounds,
			hyperPlanesQ,
			hyperPlaneNormalBounds,
			indexBins,
			halfSpaceIndices,
			hyperPlaneIndices,
			hyperPlaneNormals,
			hyperPlaneBounds
		},
		halfSpaceNormals = halfSpaceNormalsInput;
		halfSpaceBounds = halfSpaceBoundsInput;
		hyperPlaneNormals = {};
		hyperPlaneBounds = {};
		
		(* check for antipodal half-spaces (hyperplanes) *)
			halfSpaceNormalBounds = Transpose[List[halfSpaceNormals,halfSpaceBounds]];
			indexBins = GatherIndicesBy[halfSpaceNormalBounds,Function[normalBound,Sort[NRound[List[-normalBound,normalBound]]]]];
			hyperPlanesQ = NoneTrue[indexBins,Composition[GreaterThan[1],Length]];
			If[Not[hyperPlanesQ],
				hyperPlaneNormals = {};
				hyperPlaneBounds = {};
				Return[{{halfSpaceNormals,halfSpaceBounds},{hyperPlaneNormals,hyperPlaneBounds}}];
			];
		
		(*  *)
		halfSpaceIndices = Flatten[Select[indexBins,Composition[EqualTo[1],Length]]];
		hyperPlaneIndices = Part[Select[indexBins,Composition[EqualTo[2],Length]],All,1];
		hyperPlaneNormalBounds = halfSpaceNormalBounds[[hyperPlaneIndices]];
		halfSpaceNormalBounds = halfSpaceNormalBounds[[halfSpaceIndices]];

		hyperPlaneNormals = Part[hyperPlaneNormals,All,1];
		hyperPlaneBounds = Part[hyperPlaneNormalBounds,All,2];
		
		halfSpaceNormals = Part[halfSpaceNormals,All,1];
		halfSpaceBounds = Part[halfSpaceNormalBounds,All,2];
		
		Return[{{halfSpaceNormals,halfSpaceBounds},{hyperPlaneNormals,hyperPlaneBounds}}];
	]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesObjectiveVertex*)


HalfSpacesObjectiveVertex[objectiveVector_, {halfSpaceNormals_, halfSpaceBounds_}] := 
	Module[{signedBounds,lowerBound,method,vertex},
		signedBounds = ArrayPad[Transpose[{halfSpaceBounds}],{{0,0},{0,1}},-1];
		method = "Simplex";
		lowerBound = DirectedInfinity[-1];
		vertex = Quiet[LinearProgramming[-objectiveVector,halfSpaceNormals,signedBounds,lowerBound,Rule[Method,method]]];
		Return[vertex];
	]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesObjectiveBound*)


HalfSpacesObjectiveBound[objectiveVector_, {halfSpaceNormals_, halfSpaceBounds_}] := 
	Module[{vertex,bound},
		vertex = HalfSpacesObjectiveVertex[objectiveVector, {halfSpaceNormals, halfSpaceBounds}];
		If[MemberQ[vertex,Indeterminate],
			bound = Infinity;
			,
			bound = Dot[vertex,objectiveVector];
		];
		Return[bound];
	]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesFeasibleQ*)


HalfSpacesFeasibleQ[{normals_,bounds_}] := 
	Not[SameQ[LinearProgramming,Head[Quiet[HalfSpacesRandomVertex[{normals,bounds}]]]]]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesRandomVertex*)


HalfSpacesRandomVertex[{normals_,bounds_}] := HalfSpacesObjectiveVertex[RandomReal[{-1,1},VectorDimension[normals]],{normals,bounds}]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesCoordinateBounds*)


HalfSpacesCoordinateBounds[{halfSpaceNormals_,halfSpaceBounds_}] := 
	Module[{embeddingDimension,coordinateIndex,coordinateIndices},
		embeddingDimension = VectorDimension[halfSpaceNormals];
		Table[
				signValue*HalfSpacesObjectiveBound[Times[signValue,UnitVector[embeddingDimension,coordinateIndex]],{halfSpaceNormals,halfSpaceBounds}]
		,{coordinateIndex,1,embeddingDimension}
		,{signValue,{-1,+1}}
		]
	]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesAffineSpan*)


(* get the vertex coordinates for a simplex which spans the implicit region *)
HalfSpacesAffineSpan[{normals_,bounds_}] := 
	Module[{
			halfSpaces,
			embeddingDimension,
			affineSpace,
			vertices,
			newVertices,
			dimension,
			interiorPoint
		},
		halfSpaces = NChop[N[{normals,bounds}]];
		embeddingDimension = VectorDimension[normals];
		vertices = {HalfSpacesRandomVertex[halfSpaces]};
		affineSpace = VectorAffineSpan[vertices];
		Do[
			dimension = affineSpace["Dimension"];
			newVertices = NChop[Flatten[Table[
				HalfSpacesObjectiveVertex[Times[sign,objectiveVector],halfSpaces]
			,{sign,{-1,+1}},{objectiveVector,affineSpace["NullSpaceBasis"]}],1]];
			vertices = NDeleteDuplicates[Join[vertices,newVertices]];
			affineSpace = VectorAffineSpan[vertices];
			If[Equal[dimension,affineSpace["Dimension"]],Break[];];
			If[Equal[embeddingDimension,affineSpace["Dimension"]],Break[];];
		,{embeddingDimension}];
		Return[affineSpace];
	]


(* ::Subsubsection::Closed:: *)
(*HalfSpacesBoundedQ*)


HalfSpacesBoundedQ[{halfSpaceNormals_,halfSpaceBounds_}] := 
	Module[{coordinateBounds,boundedQ},
		coordinateBounds = HalfSpacesCoordinateBounds[{halfSpaceNormals,halfSpaceBounds}];
		boundedQ = FreeQ[coordinateBounds,DirectedInfinity[1]];
		Return[boundedQ];
	];		


(* ::Subtitle::Closed:: *)
(*Development*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytopeAppendPoint*)


(*
ConvexPolytopeAppendPoint[ Polytope[data_], point_ ] := 
	Module[{},
		Return[];
	]
*)


(* ::Subtitle::Closed:: *)
(*data type conversion*)


(* ::Subsubsection::Closed:: *)
(*ConvexPolytope \[Function] VectorAffineSpace*)


ConvexPolytope /: VectorAffineSpace[ConvexPolytope[p_]] := VectorAffineSpace[p[["VertexIndices",1]],p["AffineSpaceBasis"]]


ConvexPolytope /: VectorSpace[ConvexPolytope[p_]] := VectorSpace[p["AffineSpaceBasis"]]


(* ::Subtitle::Closed:: *)
(*End*)


End[];


EndPackage[];
