(* ::Package:: *)

(* ::Title:: *)
(*List`*)


BeginPackage["List`"]


(* ::Subtitle:: *)
(*Names*)


Parts;
ListIndices;
FirstMaximumIndex;
FirstMinimumIndex;
ListDuplicates;
ListDuplicatesBy;
ListNonDuplicates;
ListNonDuplicatesBy;
IndexSubsetAdjacencyMatrix;
IndexSubsetAdjacencyMatrixSparse;
IndexSubsetMemberships;
IntersectionIndices;
GatherIndices;
GatherIndicesBy;
GatherByList;
ListPlusMinus;
ListEqualTo;
ListUnequalTo;
ListGreaterThan;
ListGreaterEqualThan;
ListLessThan;
ListLessEqualThan;
ListEqualToBoole;
ListUnequalToBoole;
ListGreaterThanBoole;
ListGreaterEqualThanBoole;
ListLessThanBoole;
ListLessEqualThanBoole;
TransposeArray;


(* ::Subtitle:: *)
(*Definitions*)


Begin["`Private`"]


(* ::Subsection::Closed:: *)
(*part*)


(* ::Subsubsection::Closed:: *)
(*Parts*)


Parts::usage = 
"Parts[list,indices] returns the parts of list corresponding to the indices.
Parts[list] represents an operator which can be applied to an indicie or (nested) list of indicies.";


Parts[list_List, index_Integer] := Part[list,index]
Parts[list_List, indices_?VectorQ] := Part[list,indices]
Parts[list_List, indices_?ArrayQ] := Map[Part[list,#]&,indices,{-2}]
Parts[list_List, indices_] := Map[Parts[list,#]&,indices]
Parts[list_List][indices_] := Parts[list, indices]


(* ::Subsection:: *)
(*index*)


(* ::Subsubsection::Closed:: *)
(*ListIndices*)


ListIndices::usage = 
"ListIndices[list] returns the indices of list.";


ListIndices[list_List] := Range[Length[list]]


(* ::Subsubsection:: *)
(*IntersectionIndices*)


IntersectionIndices::usage = 
"IntersectionIndices[x,y] returns the indices of A for which A\[LeftDoubleBracket]i\[RightDoubleBracket]\[Element]B.";


IntersectionIndices//ClearAll;
IntersectionIndices[x_,y_] := 
	Module[{z,zxN,zxI,xI},
		z = Intersection[x,y];
		zxI = Take[Values[PositionIndex[Join[z,x]]],Length[z]];
		xI = Subtract[Part[zxI,All,2],Length[z]];
		Return[xI];
	]


(* ::Subsubsection:: *)
(*FirstMinimumIndex*)


FirstMinimumIndex[values_List] := First[Ordering[values,1]]


(* ::Subsubsection:: *)
(*FirstMaximumIndex*)


FirstMaximumIndex[values_List] := First[Ordering[Minus[values],1]]


(* ::Subsection::Closed:: *)
(*index subsets*)


(* ::Subsubsection::Closed:: *)
(*IndexSubsetAdjacencyMatrix*)


IndexSubsetMemberships::usage = 
"IndexSubsetMemberships[subsets] returns the list(i) = {j : i\[Element]subset[[j]]}.";


IndexSubsetAdjacencyMatrix[indexSubsets_,maxIndex_:0] := 
	Module[{adjacencyMatrix,max},
		max = Max[maxIndex,Max[indexSubsets]];
		adjacencyMatrix = ConstantArray[0,{max,Length[indexSubsets]}];
		Do[
			adjacencyMatrix[[indexSubsets[[subsetIndex]],subsetIndex]] = 1;
		,{subsetIndex,Length[indexSubsets]}];
		Return[adjacencyMatrix];
	];


(* ::Subsubsection::Closed:: *)
(*IndexSubsetAdjacencyMatrixSparse*)


IndexSubsetAdjacencyMatrixSparse::usage = 
"IndexSubsetAdjacencyMatrixSparse[subsets] returns a sparse adjacency matrix where M(i,j)=1 if subset[[j]] contains the index i.";


IndexSubsetAdjacencyMatrixSparse[pointIndexSubsets_] := 
	SparseArray[Flatten[MapIndexed[Composition[Tuples,List],pointIndexSubsets],1] -> 1]


(* ::Subsubsection::Closed:: *)
(*IndexSubsetMemberships*)


IndexSubsetMemberships::usage = 
"IndexSubsetMemberships[subsets] returns the list(i) = {j : i \[Element] subsets[[j]]}.";


IndexSubsetMemberships[pointIndexSubsets_] := 
	Module[{pointSubsetAdjacencyMatrix,pointSubsetIndices,subsetIndices},
		subsetIndices = ListIndices[pointIndexSubsets];
		pointSubsetAdjacencyMatrix = IndexSubsetAdjacencyMatrixSparse[pointIndexSubsets];
		pointSubsetIndices = Map[Function[pointSubsetAdj,Pick[subsetIndices,pointSubsetAdj,1]],pointSubsetAdjacencyMatrix];
		Return[pointSubsetIndices];
	]


(* ::Subsection::Closed:: *)
(*duplicates*)


(* ::Subsubsection::Closed:: *)
(*ListDuplicates*)


ListDuplicates::usage = 
"SelectDuplicates[list] returns the elements of list which appear more than once.";


ListDuplicates[list_] := 
	Module[{valuesUnique,valueMultiplicities,duplicateValues},
		{valuesUnique,valueMultiplicities} = Transpose[Tally[list]];
		duplicateValues = Pick[valuesUnique,Unitize[Subtract[valueMultiplicities,1]],1];
		Return[duplicateValues];
	]


(* ::Subsubsection::Closed:: *)
(*ListDuplicatesBy*)


ListDuplicatesBy::usage = 
"ListDuplicatesBy[list, f] returns the elements of list whose f-values appear more than once.";


ListDuplicatesBy[list_, f_] := 
	Module[{valuesUnique,valueMultiplicities,duplicateValues},
		{valuesUnique,valueMultiplicities} = Transpose[Tally[list,Function[SameQ[f[#1],f[#2]]]]];
		duplicateValues = Pick[valuesUnique,Unitize[Subtract[valueMultiplicities,1]],1];
		Return[duplicateValues];
	]


(* ::Subsection::Closed:: *)
(*non-duplicates*)


(* ::Subsubsection::Closed:: *)
(*ListNonDuplicates*)


ListNonDuplicates::usage = 
"ListDuplicates[list] returns the elements of list which appear only once.";


ListNonDuplicates[list_] := 
	Module[{tally,valuesUnique,valueMultiplicities,duplicateValues},
		tally = Tally[list];
		{valuesUnique,valueMultiplicities} = Transpose[tally];
		duplicateValues = Pick[valuesUnique,valueMultiplicities,1];
		Return[duplicateValues];
	]


(* ::Subsubsection::Closed:: *)
(*ListNonDuplicatesBy*)


ListNonDuplicatesBy::usage = 
"ListDuplicatesBy[list, f] returns the elements of list whose f-values appear only once.";


ListNonDuplicatesBy[list_, f_] := 
	Module[{valuesUnique,valueMultiplicities,duplicateValues},
		{valuesUnique,valueMultiplicities} = Transpose[Tally[list,Function[SameQ[f[#1],f[#2]]]]];
		duplicateValues = Pick[valuesUnique,valueMultiplicities,1];
		Return[duplicateValues];
	]


(* ::Subsection:: *)
(*gather*)


(* ::Subsubsection::Closed:: *)
(*GatherIndices*)


GatherIndices::usage = 
"GatherIndices[list] gathers the indices of list by the values of their corresponding elements list.";


GatherIndices[list_] := Values[PositionIndex[list]]


(* ::Subsubsection::Closed:: *)
(*GatherIndicesBy*)


GatherIndicesBy::usage = 
	"GatherIndices[list, f] gathers the indices of list by the f-values of their corresponding elements list.";


GatherIndicesBy[list_, f_] := Values[PositionIndex[Map[f,list]]]


(* ::Subsubsection::Closed:: *)
(*GatherByList*)


GatherByList::usage = 
	"GatherByList[elements, values] gathers elements into lists according to their corresponding values.";


GatherByList[elements_, values_] := Parts[elements,GatherIndices[values]]


(* ::Subsection:: *)
(*operations*)


(* ::Subsubsection::Closed:: *)
(*ListPlusMinus*)


ListPlusMinus::usage = 
	"ListPlusMinus[x] returns the sorted list ";


ListPlusMinus[x_] := NumericalSort[{-x,+x}]


(* ::Subsection:: *)
(*compare list to value*)


(* ::Subsubsection::Closed:: *)
(*ListEqualTo*)


ListEqualTo::usage = 
	"ListEqualTo[list, value] returns the list {(element==value)\[Element]{True,False}} for each element in list.";


ListEqualTo[list_, value_] := NonPositive[Unitize[Subtract[list,value]]]


(* ::Subsubsection::Closed:: *)
(*ListUnequalTo*)


ListUnequalTo::usage = 
	"ListUnequalTo[list, value] returns the list {(element \[NotEqual] value)\[Element]{True,False}} for each element in list.";


ListUnequalTo[list_, value_] := Positive[Unitize[Subtract[list,value]]]


(* ::Subsubsection::Closed:: *)
(*ListLessThan*)


ListLessThan::usage = 
	"ListEqualTo[list, value] returns the list {(element < value)\[Element]{True,False}} for each element in list.";


ListLessThan[list_, value_] := Negative[Subtract[list,value]]


(* ::Subsubsection::Closed:: *)
(*ListLessEqualThan*)


ListLessEqualThan::usage = 
	"ListLessEqualThan[list, value] returns the list {(element \[LessEqual] value)\[Element]{True,False}} for each element in list.";


ListLessEqualThan[list_, value_] := NonPositive[Subtract[list,value]]


(* ::Subsubsection::Closed:: *)
(*ListGreaterThan*)


ListGreaterThan::usage = 
	"ListGreaterThan[list, value] returns the list {(element > value)\[Element]{True,False}} for each element in list.";


ListGreaterThan[list_, value_] := Positive[Subtract[list,value]]


(* ::Subsubsection::Closed:: *)
(*ListGreaterEqualThan*)


ListGreaterEqualThan::usage = 
	"ListGreaterEqualThan[list, value] returns the list {(element \[GreaterEqual] value)\[Element]{True,False}} for each element in list.";


ListGreaterEqualThan[list_, value_] := NonNegative[Subtract[list,value]]


(* ::Subsection::Closed:: *)
(*list compare (boole)*)


(* ::Subsubsection::Closed:: *)
(*ListEqualToBoole*)


ListEqualToBoole::usage = 
	"ListEqualToBoole[list, value] returns the list {(element \[Equal] value)\[Element]{1,0}} for each element in list.";


ListEqualToBoole[list_, value_] := Subtract[1,Unitize[Subtract[list,value]]]


(* ::Subsubsection::Closed:: *)
(*ListUnequalToBoole*)


ListUnequalToBoole::usage = 
	"ListUnequalToBoole[list, value] returns the list {(element \[NotEqual] value)\[Element]{1,0}} for each element in list.";


ListUnequalToBoole[list_, value_] := Unitize[Subtract[list,value]]


(* ::Subsubsection::Closed:: *)
(*ListLessThanBoole*)


ListLessThanBoole::usage = 
	"ListLessThanBoole[list, value] returns the list {(element < value)\[Element]{1,0}} for each element in list."


ListLessThanBoole[list_, value_] := Ramp[Sign[Subtract[value,list]]]


(* ::Subsubsection::Closed:: *)
(*ListLessEqualThanBoole*)


ListLessEqualThanBoole::usage = 
	"ListLessEqualThanBoole[list, value] returns the list {(element \[LessEqual] value)\[Element]{1,0}} for each element in list.";


ListLessEqualThanBoole[list_, value_] := UnitStep[Subtract[value,list]]


(* ::Subsubsection::Closed:: *)
(*ListGreaterThanBoole*)


ListGreaterThanBoole::usage = 
	"ListGreaterThanBoole[list, value] returns the list {(element > value)\[Element]{1,0}} for each element in list.";


ListGreaterThanBoole[list_, value_] := Ramp[Sign[Subtract[list,value]]]


(* ::Subsubsection::Closed:: *)
(*ListGreaterEqualThanBoole*)


ListGreaterEqualThanBoole::usage = 
	"ListGreaterEqualThanBoole[list, value] returns the list {(element \[GreaterEqual] value)\[Element]{1,0}} for each element in list.";


ListGreaterEqualThanBoole[list_, value_] := UnitStep[Subtract[list,value]]


(* ::Subtitle:: *)
(*End*)


(* ::Subsubsection::Closed:: *)
(*TransposeArray*)


TransposeArray[array_] := Transpose[array,TwoWayRule[1,ArrayDepth[array]]]



End[]


EndPackage[]
